/***************************************************************************
 *   nipper - The network infrastructure parser                            *
 *   Copyright (C) 2006 - 2007 by Ian Ventura-Whiting (Fizz)               *
 *   fizz@titania.co.uk                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program. If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************

The following files should be provided with this source code:

* INSTALL   - The compilation and installation guide
* LICENCE   - The licence covering all the files
* Changelog - A listing of all source code changes
* Readme    - Information about Cisco Parse


Program Overview
~~~~~~~~~~~~~~~~

Nipper is a Network Infrastructure Configuration Parser. Nipper takes
a network infrastructure device configuration, processes the file and
details security-related issues with the configuration together with
detailed recommendations. Nipper was previous known as CiscoParse.

Nipper currently supports Cisco switches (IOS), Cisco Routers (IOS),
Cisco Firewalls (PIX/ASA/FWSM) and Juniper NetScreen (ScreenOS). Output
is in HTML, Latex, XML and Text. Encrypted passwords can be output to a
John-the-Ripper file for strength testing.


Program Usage
~~~~~~~~~~~~~

For parsing a Cisco router configuration file called router.txt and
outputting as a text file called report.txt:

$ nipper --input=router.txt --output=report.txt --output-format=text

For parsing a PIX firewall configuration file called firewall.txt and
outputing as a HTML file called report.html:

$ nipper --pix --input=firewall.txt --output=report.html

For more options, view the nipper help:

$ nipper --help


Program Code
~~~~~~~~~~~~

This program code is split into five distinct parts:

  1) Program parameters and initialisation.
  2) Process the configuration file.
  3) Report Pre-processing (defaults...).
  4) Reporting and output
  5) Cleanup

*/


// Standard Includes...
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>

// Global Constants...
const char *program_banner = "                     _                           ____\n"
                             "               _ __ (_)_ __  _ __   ___ _ __    / ->/|\n"
                             "              | '_ \\| | '_ \\| '_ \\ / _ \\ '__|  /<-_/ |\n"
                             "              | | | | | |_) | |_) |  __/ |     |   | /\n"
                             "              |_| |_|_| .__/| .__/ \\___|_|     |___|/\n"
                             "                      |_|   |_|\n\n"
                             "                           Version 0.10.5\n"
                             "               by Ian Ventura-Whiting and David Croft\n"
                             "                   http://nipper.sourceforge.net\n";
const char *program_version_banner = "nipper version 0.10.5\nby Ian Ventura-Whiting and David Croft\nhttp://nipper.sourceforge.net";
const char *program_version = "0.10.5";

// Common structure include...
#include "nipper-struct.c"

// Common functions include...
#include "nipper-common.c"

// Common report include...
#include "report-appendix.c"
#include "report-text.c"
#include "report-functions.c"
#include "report-vulns.c"

// Processing IOS includes...
#include "IOS/process-aaa.c"
#include "IOS/process-access-list.c"
#include "IOS/process-banner.c"
#include "IOS/process-cef.c"
#include "IOS/process-clock.c"
#include "IOS/process-dns.c"
#include "IOS/process-enable.c"
#include "IOS/process-ftp.c"
#include "IOS/process-general.c"
#include "IOS/process-http.c"
#include "IOS/process-interface.c"
#include "IOS/process-keychain.c"
#include "IOS/process-line.c"
#include "IOS/process-logging.c"
#include "IOS/process-nat.c"
#include "IOS/process-ntp.c"
#include "IOS/process-route-map.c"
#include "IOS/process-router.c"
#include "IOS/process-snmp.c"
#include "IOS/process-ssh.c"
#include "IOS/process-tacacs.c"
#include "IOS/process-username.c"
#include "IOS/process-vtp.c"

// Report IOS includes...
#include "IOS/report-passwords.c"
#include "IOS/report-timeout.c"
#include "IOS/report-debug.c"
#include "IOS/report.c"

// Processing NMP includes...
#include "NMP/process-general.c"
#include "NMP/process-permit.c"
#include "NMP/process-port.c"
#include "NMP/process-snmp.c"
#include "NMP/process-spantree.c"
#include "NMP/process-vlan.c"
#include "NMP/process-cdp.c"

// Report NMP includes...
#include "NMP/report-passwords.c"
#include "NMP/report-debug.c"
#include "NMP/report.c"

// Processing PIX includes...
#include "PIX/process-access-list.c"
#include "PIX/process-enable.c"
#include "PIX/process-general.c"
#include "PIX/process-icmp.c"
#include "PIX/process-interface.c"
#include "PIX/process-nat.c"
#include "PIX/process-objects.c"
#include "PIX/process-protocols.c"
#include "PIX/process-route.c"
#include "PIX/process-ssh.c"

// Report PIX includes...
#include "PIX/report-debug.c"
#include "PIX/report-access-list.c"
#include "PIX/report.c"

// Processing CSS includes...
#include "CSS/process-access-list.c"
#include "CSS/process-restrict.c"
#include "CSS/process-snmp.c"

// Report CSS includes...
#include "CSS/report-debug.c"
#include "CSS/report.c"

// Processing ScreenOS includes...
#include "ScreenOS/process-address.c"
#include "ScreenOS/process-admin.c"
#include "ScreenOS/process-auth-server.c"
#include "ScreenOS/process-interface.c"
#include "ScreenOS/process-policy.c"
#include "ScreenOS/process-ssh.c"

// Report SOS includes...
#include "ScreenOS/report-debug.c"
#include "ScreenOS/report.c"

// Processing includes...
#include "nipper-defaults.c"
#include "nipper-namemappings.c"

// Cleanup Structures...
#include "IOS/cleanup.c"
#include "NMP/cleanup.c"
#include "PIX/cleanup.c"
#include "CSS/cleanup.c"
#include "ScreenOS/cleanup.c"


// ----------------------------------------------------
// PART 1: Program parameters & initialization

int main(int argc, char *argv[])
{
	// Program variables...
	int showUsage = false;				// Should usage information be shown? (default to false)
	int showVersion = false;
	int inputFile = 0;					// Argument that contains input file, if specified.
	int argLoop = 0;					// Arguement loop variable
	int setting = 0;
	int tempInt = 0;
	int tempInt2 = 0;
	char line[1024];					// Line from configuration file
	char tempCommand[256];				// Retrival of Config remotely - Command to execute
	struct stat *fileStats = 0;
	struct ciscoCommand command;
	struct debugAbbrev *debugAbPointer = 0;

	// nipper Global Configuration...
	struct nipperConfig nipper;
	memset(&nipper, 0, sizeof(struct nipperConfig));

	// nipper Program Option Defaults...
	nipper.deviceType = type_ios_router;
	nipper.deviceUse = use_all;
	strcpy(nipper.paperSize, "a4paper");
	strcpy(nipper.documentClass, "article");
	strcpy(nipper.remoteSnmp, "private");
	strcpy(nipper.remoteTftp, "/tftproot/");
	strcpy(nipper.companyName, "Nipper");
	nipper.outputFormat = output_html;
	nipper.debugMode = false;
	nipper.linesnotprocessed = false;
	nipper.abs = true;
	nipper.appendix = true;
	nipper.loggingLevels = unconfigured;
	nipper.timeZones = unconfigured;
	nipper.glossary = true;
	nipper.nipperVersion = true;
	nipper.commonPorts = true;
	nipper.fullReport = true;
	nipper.debugAbbrev = false;
	nipper.expandACL = false;
	nipper.names = true;
	nipper.passwords = true;
	nipper.localSave = 0;
	nipper.denyLog = true;
	nipper.strictACL = false;
	nipper.oldCisco = false;
	nipper.passwordLength = 8;
	nipper.passwordUppers = false;
	nipper.passwordLowers = false;
	nipper.passwordEither = true;
	nipper.passwordNumbers = true;
	nipper.passwordSpecials = false;
	nipper.connectionTimeout = 600;
	nipper.maxSize = sizeof(line);
	nipper.tableNum = 1;

	// Combined Configuration Variables...
	struct johnPassword *johnPointer;

	// Init program configuration...
	memset(tempCommand, 0, sizeof(tempCommand));

	// Get program parameters
	for (argLoop = 1; argLoop < argc; argLoop++)
	{
		// Output format specified
		if ((strncmp("--output-format=", argv[argLoop], 16) == 0) || (strncmp("--report-format=", argv[argLoop], 16) == 0))
		{
			if (strcmp("text", argv[argLoop]+16) == 0)
				nipper.outputFormat = output_text;
			else if (strcmp("html", argv[argLoop]+16) == 0)
				nipper.outputFormat = output_html;
			else if (strcmp("latex", argv[argLoop]+16) == 0)
				nipper.outputFormat = output_latex;
			else if (strcmp("xml", argv[argLoop]+16) == 0)
				nipper.outputFormat = output_xml;
			else
				showUsage = true;
		}
		else if (strncmp("--text", argv[argLoop], 6) == 0)
			nipper.outputFormat = output_text;
		else if (strncmp("--html", argv[argLoop], 6) == 0)
			nipper.outputFormat = output_html;
		else if (strncmp("--latex", argv[argLoop], 7) == 0)
			nipper.outputFormat = output_latex;
		else if (strncmp("--xml", argv[argLoop], 5) == 0)
			nipper.outputFormat = output_xml;

		// Input file (defaults to stdin)
		else if ((strncmp("--input=", argv[argLoop], 8) == 0) && (strlen(argv[argLoop]) > 8))
			inputFile = argLoop;

		// Output file (defaults to stdout)
		else if (((strncmp("--output=", argv[argLoop], 9) == 0) || (strncmp("--report=", argv[argLoop], 9) == 0)) && (strlen(argv[argLoop]) > 9))
			strncpy(nipper.outputFile, argv[argLoop]+9, sizeof(nipper.outputFile) - 1);

		// Password output
		else if (strcmp("--no-passwords", argv[argLoop]) == 0)
			nipper.passwords = false;

		// Do not perform a security audit
		else if (strcmp("--no-audit", argv[argLoop]) == 0)
			nipper.fullReport = false;

		// Expand ACL object definitions
		else if (strcmp("--expand-acl", argv[argLoop]) == 0)
			nipper.expandACL = true;

		// Debug output
		else if (strcmp("--debug", argv[argLoop]) == 0)
			nipper.debugMode = true;
		else if (strcmp("--debug-lnp", argv[argLoop]) == 0)
			nipper.linesnotprocessed = true;

		// Router
		else if ((strcmp("--ios-router", argv[argLoop]) == 0) || (strcmp("--ios", argv[argLoop]) == 0))
			nipper.deviceType = type_ios_router;

		// Cisco Router
		else if (strcmp("--ios-switch", argv[argLoop]) == 0)
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_ios_switch;
			else
				showUsage = true;
		}

		// IOS-based Catalyst
		else if (strcmp("--ios-catalyst", argv[argLoop]) == 0)
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_ios_catalyst;
			else
				showUsage = true;
		}

		// Cisco Firewall
		else if ((strcmp("--pix", argv[argLoop]) == 0) || (strcmp("--asa", argv[argLoop]) == 0) || (strcmp("--fwsm", argv[argLoop]) == 0))
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_pix_firewall;
			else
				showUsage = true;
		}

		// Cisco NMP/CatOS Catalyst
		else if ((strcmp("--catos", argv[argLoop]) == 0) || (strcmp("--nmp", argv[argLoop]) == 0))
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_nmp_catalyst;
			else
				showUsage = true;
		}

		// Cisco CSS
		else if (strcmp("--css", argv[argLoop]) == 0)
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_css_filter;
			else
				showUsage = true;
		}

		// Juniper ScreenOS
		else if ((strcmp("--netscreen", argv[argLoop]) == 0) || (strcmp("--screenos", argv[argLoop]) == 0))
		{
			if (nipper.deviceType == type_ios_router)
				nipper.deviceType = type_sos_firewall;
			else
				showUsage = true;
		}

		// Set Timeout
		else if (strncmp("--timeout=", argv[argLoop], 10) == 0)
		{
			if (atoi(argv[argLoop] + 10) != 0)
				nipper.connectionTimeout = atoi(argv[argLoop] + 10);
		}

		// Dictionary supplied
		else if (strncmp("--stylesheet=", argv[argLoop], 13) == 0)
			nipper.cssFile = fopen(argv[argLoop] + 13, "r");

		// Company name specified
		else if (strncmp("--company-name=", argv[argLoop], 15) == 0)
			strncpy(nipper.companyName, argv[argLoop] + 15, sizeof(nipper.companyName) - 1);

		// Company name specified
		else if (strncmp("--device-name=", argv[argLoop], 14) == 0)
			strncpy(nipper.hostname, argv[argLoop] + 14, sizeof(nipper.hostname) - 1);

		// Password Length
		else if (strncmp("--pass-length=", argv[argLoop], 14) == 0)
			nipper.passwordLength = atoi(argv[argLoop] + 14);

		// Password Uppers
		else if (strcmp("--pass-uppers=yes", argv[argLoop]) == 0)
			nipper.passwordUppers = true;
		else if (strcmp("--pass-uppers=no", argv[argLoop]) == 0)
			nipper.passwordUppers = false;

		// Password Lowers
		else if (strcmp("--pass-lowers=yes", argv[argLoop]) == 0)
			nipper.passwordLowers = true;
		else if (strcmp("--pass-lowers=no", argv[argLoop]) == 0)
			nipper.passwordLowers = false;

		// Password Either
		else if (strcmp("--pass-either=yes", argv[argLoop]) == 0)
			nipper.passwordEither = true;
		else if (strcmp("--pass-either=no", argv[argLoop]) == 0)
			nipper.passwordEither = false;

		// Password Numbers
		else if (strcmp("--pass-numbers=yes", argv[argLoop]) == 0)
			nipper.passwordNumbers = true;
		else if (strcmp("--pass-numbers=no", argv[argLoop]) == 0)
			nipper.passwordNumbers = false;

		// Password Specials
		else if (strcmp("--pass-specials=yes", argv[argLoop]) == 0)
			nipper.passwordSpecials = true;
		else if (strcmp("--pass-specials=no", argv[argLoop]) == 0)
			nipper.passwordSpecials = false;

		// Deny all and log filtering
		else if (strcmp("--no-deny-log", argv[argLoop]) == 0)
			nipper.denyLog = false;

		// Strict ACL
		else if (strcmp("--strict-source", argv[argLoop]) == 0)
			nipper.strictACL = true;

		// Latex paper size
		else if (strncmp("--paper=", argv[argLoop], 8) == 0)
			strncpy(nipper.paperSize, argv[argLoop]+8, sizeof(nipper.paperSize) - 1);

		// Latex document class
		else if (strncmp("--documentclass=", argv[argLoop], 16) == 0)
			strncpy(nipper.documentClass, argv[argLoop]+16, sizeof(nipper.documentClass) - 1);

		// Appendix
		else if (strcmp("--no-appendix", argv[argLoop]) == 0)
			nipper.appendix = false;

		// Glossary
		else if (strcmp("--no-glossary", argv[argLoop]) == 0)
			nipper.glossary = false;

		// Common Ports
		else if (strcmp("--no-ports", argv[argLoop]) == 0)
			nipper.commonPorts = false;

		// Nipper Version Appendix
		else if (strcmp("--no-version", argv[argLoop]) == 0)
			nipper.nipperVersion = false;

		// Edge device
		else if (strcmp("--edge", argv[argLoop]) == 0)
		{
			if (nipper.deviceUse == use_all)
				nipper.deviceUse = use_edge;
			else
				showUsage = true;
		}

		// Internal device
		else if (strcmp("--internal", argv[argLoop]) == 0)
		{
			if (nipper.deviceUse == use_all)
				nipper.deviceUse = use_internal;
			else
				showUsage = true;
		}

		// Logging Levels Appendix
		else if (strcmp("--no-logging", argv[argLoop]) == 0)
			nipper.loggingLevels = false;

		// Time Zone Appendix
		else if (strcmp("--no-timezones", argv[argLoop]) == 0)
			nipper.timeZones = false;

		// Abbreviations
		else if (strcmp("--no-abbreviations", argv[argLoop]) == 0)
			nipper.abs = false;

		// Show host / network names instead of IP addresses (where possible)
		else if (strcmp("--no-names", argv[argLoop]) == 0)
			nipper.names = false;

		// Show host / network names instead of IP addresses (where possible)
		else if (strcmp("--debug-abbreviations", argv[argLoop]) == 0)
			nipper.debugAbbrev = true;

		// John the ripper output
		else if ((strncmp("--john=", argv[argLoop], 7) == 0) && (strlen(argv[argLoop]) > 7))
			strncpy(nipper.johnFile, argv[argLoop]+7, sizeof(nipper.johnFile) - 1);

		// Dictionary supplied
		else if (strncmp("--dictionary=", argv[argLoop], 13) == 0)
			nipper.dictionaryFile = fopen(argv[argLoop] + 13, "r");

#if !defined(__win32__)
		// Remote Cisco Device IP
		else if (strncmp("--cisco-ip=", argv[argLoop], 11) == 0)
			strncpy(nipper.remoteIP, argv[argLoop] + 11, sizeof(nipper.remoteIP) - 1);

		// Remote Cisco Device IP (Local)
		else if (strncmp("--local-ip=", argv[argLoop], 11) == 0)
			strncpy(nipper.localIP, argv[argLoop] + 11, sizeof(nipper.localIP) - 1);

		// Remote Cisco Device SNMP read/write community
		else if (strncmp("--snmp=", argv[argLoop], 7) == 0)
			strncpy(nipper.remoteSnmp, argv[argLoop] + 7, sizeof(nipper.remoteSnmp) - 1);

		// Local TFTP server to use
		else if (strncmp("--tftproot=", argv[argLoop], 11) == 0)
			strncpy(nipper.remoteTftp, argv[argLoop] + 11, sizeof(nipper.remoteTftp) - 1);

		// Local Cisco Config file to save to
		else if (strncmp("--cisco-file=", argv[argLoop], 13) == 0)
			nipper.localSave = argLoop;

		// Cisco retrival method (old or new)
		else if (strcmp("--cisco=old", argv[argLoop]) == 0)
			nipper.oldCisco = true;
		else if (strcmp("--cisco=new", argv[argLoop]) == 0)
			nipper.oldCisco = false;
#endif

		// Version
		else if (strcmp("--version", argv[argLoop]) == 0)
			showVersion = true;

		// If all else fails, display the help
		else
			showUsage = true;
	}

	// Show help?
	if (showUsage == true)
	{
		printf("%s%s\n%s", COL_BLUE, program_banner, RESET);
		printf("Nipper is a  Network Infrastructure  Configuration Parser.  Nipper takes\n");
		printf("a network infrastructure  device configuration,  processes the  file and\n");
		printf("details security-related  issues with detailed  recommendations.  Nipper\n");
		printf("was previous known as CiscoParse.\n\n");

		printf("Nipper  currently  supports  Cisco switches (IOS),  Cisco Routers (IOS),\n");
		printf("Cisco Catalyst devices (IOS/CatOS/NMP),  Cisco Firewalls (PIX/ASA/FWSM),\n");
        printf("Cisco Content  Services Switch (CSS)  and Juniper  NetScreen (ScreenOS).\n");
		printf("Output from Nipper is in HTML, Latex, XML and Text. Additionally, Nipper\n");
		printf("will reverse any Cisco  type-7 passwords, any other  encrypted passwords\n");
		printf("can be output to a John-the-Ripper file for further testing.\n");

		printf("By default, input is retrieved from stdin and is output (in HTML format)\n");
		printf("to stdout.\n\n");

		printf("%sCommand:%s\n\n", COL_BLUE, RESET);
		printf("    %s%s [Options]%s\n\n", COL_GREEN, argv[0], RESET);

		printf("%sGeneral Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--help%s\n", COL_GREEN, RESET);
		printf("    Displays this command line help and overrides all other options.\n\n");
		printf("    %s--input=<file>%s\n", COL_GREEN, RESET);
		printf("    Specifies a device configuration file to be processed.\n\n");
		printf("    %s--output=<file> | --report=<file>%s\n", COL_GREEN, RESET);
		printf("    Specified an output file for the report.\n\n");
		printf("    %s--version%s\n", COL_GREEN, RESET);
		printf("    Displays the program version.\n\n");

		printf("%sNetwork Infrastructure Device Type Options:%s\n\n", COL_BLUE, RESET);
		printf("    Different types  of network  infrastructure  devices have  different\n");
		printf("    configuration file structures and default options.\n\n");
		printf("    %s--pix | --asa | --fwsm%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a PIX/ASA/FWSM-based Cisco firewall.\n\n");
		printf("    %s--ios | --ios-router%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a  IOS-based Cisco router.  This is the\n");
		printf("    default device type.\n\n");
		printf("    %s--ios-switch%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a IOS-based Cisco switch.\n\n");
		printf("    %s--ios-catalyst%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a IOS-based Cisco catalyst.\n\n");
		printf("    %s--catos | --nmp%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a CatOS- or NMP-based Cisco Catalyst.\n\n");
		printf("    %s--css%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a Cisco Content Services Switch.\n\n");
		printf("    %s--netscreen | --screenos%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a ScreenOS-based Juniper Netscreen\n");
		printf("    firewall device.\n\n");

		printf("%sNetwork Infrastructure Device Usage Options:%s\n\n", COL_BLUE, RESET);
		printf("    Different  types of  network infrastructure  devices have  different\n");
		printf("    uses within  a network  infrastructure.  By default  nipper will not\n");
		printf("    treat the device as if it has a particular use.\n\n");
		printf("    %s--edge%s\n", COL_GREEN, RESET);
		printf("    Treat the device as a network infrastructure edge device.\n\n");
		printf("    %s--internal%s\n", COL_GREEN, RESET);
		printf("    Treat the device as an internal network infrastructure device.\n\n");

		printf("%sGeneral Report Options:%s\n\n", COL_BLUE, RESET);
		printf("    By  default a  report will  include a  security audit  section which\n");
		printf("    includes any identified  security-related issues,  their impact  and\n");
		printf("    recommendations.   Additionally,   it   will    include   a   device\n");
		printf("    configuration section and an appendix section.\n\n");
		printf("    %s--output-format=<format> | --report-format=<format>%s\n", COL_GREEN, RESET);
		printf("    Defines the output  format produced  by nipper.  Valid formats  are:\n");
		printf("    \"html\", \"latex\", \"text\" and \"xml\". The default is html.\n\n");
		printf("    %s--no-audit%s\n", COL_GREEN, RESET);
		printf("    This option disables the security audit of the device configuration.\n\n");
		printf("    %s--no-names%s\n", COL_GREEN, RESET);
		printf("    Convert host / network names to IP addresses where possible.\n\n");
		printf("    %s--expand-acl%s\n", COL_GREEN, RESET);
		printf("    Expand an Access Control Lists / Filtering  object definitions. This\n");
		printf("    option will show the reality of a filtering policy.\n\n");
		printf("    %s--device-name=<name>%s\n", COL_GREEN, RESET);
		printf("    If a device configuration  file does not  contain a device name,  it\n");
		printf("    can be specified using this option.\n\n");
		printf("    %s--company-name=<name>%s\n", COL_GREEN, RESET);
		printf("    Replace the 'nipper recommends...' in the report output to something\n");
		printf("    else. So for the Acme company, it would read, 'Acme recommends...'.\n\n");

		printf("%sAppendix Report Options:%s\n\n", COL_BLUE, RESET);
		printf("    These options allow for the  customisation of  the reports  appendix\n");
		printf("    section. However, not all  these report sections may  appear in your\n");
		printf("    report.  Appendix   sections   are   custom   generated   for   each\n");
		printf("    configuration.\n\n");
		printf("    %s--no-appendix%s\n", COL_GREEN, RESET);
		printf("    Disables  the output  of  the entire  report  appendix  section.  By\n");
		printf("    default an appendix section is generated and added to the report.\n\n");
		printf("    %s--no-abbreviations%s\n", COL_GREEN, RESET);
		printf("    Disables the output of an abbreviations section in the appendix.\n\n");
		//printf("    %s--no-glossary%s\n", COL_GREEN, RESET);
		//printf("    Disables the output of a glossary section in the appendix.\n\n");
		printf("    %s--no-logging%s\n", COL_GREEN, RESET);
		printf("    Disables the output of the logging levels section in the appendix.\n\n");
		printf("    %s--no-timezones%s\n", COL_GREEN, RESET);
		printf("    Disables the output of the time zones section in the appendix.\n\n");
		printf("    %s--no-ports%s\n", COL_GREEN, RESET);
		printf("    Disables the output of a common ports section in the appendix.\n\n");
		printf("    %s--no-version%s\n", COL_GREEN, RESET);
		printf("    Disables the output of the nipper version section in the appendix.\n\n");

		printf("%sReport Format Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--stylesheet=<file>%s\n", COL_GREEN, RESET);
		printf("    Supply an external Cascading Style Sheet (CSS) for use with the HTML\n");
		printf("    -based reports.  If this  option is not  supplied a  default  CSS is\n");
		printf("    embeded. The  supplied CSS  file will be  embeded within  the report\n");
		printf("    output.\n\n");
		printf("    %s--paper=<Latex paper size>%s\n", COL_GREEN, RESET);
		printf("    This option can be used to configure the Latex document  paper size.\n");
		printf("    The   default  paper   size  is  '%s'.   Refer  to   the  Latex\n", nipper.paperSize);
		printf("    documentation for other settings.\n\n");
		printf("    %s--documentclass=<Latex document class>%s\n", COL_GREEN, RESET);
		printf("    This option can be used to configure  the Latex document  class. The\n");
		printf("    default  document  class   is   '%s'.   Refer   to   the  Latex\n", nipper.documentClass);
		printf("    documentation for other settings.\n\n");

		printf("%sSecurity Audit Network Filtering Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--strict-source%s\n", COL_GREEN, RESET);
		printf("    Check for filtering from any source.\n\n");
		printf("    %s--no-deny-log%s\n", COL_GREEN, RESET);
		printf("    Prevents checking for filtering not ending with a deny all and log.\n\n");

		printf("%sSecurity Audit Password Options:%s\n\n", COL_BLUE, RESET);
		printf("    Password strength and  dictionary-based testing is  performed on all\n");
		printf("    authentication systems.  Therefore, in addition  to user  passwords,\n");
		printf("    SNMP community strings and keys are also checked.\n\n");
		printf("    %s--no-passwords%s\n", COL_GREEN, RESET);
		printf("    Removes passwords  from the  output. However,  passwords are  always\n");
		printf("    checked against defaults and for password strength.\n\n");
		printf("    %s--john=<file>%s\n", COL_GREEN, RESET);
		printf("    Output the Cisco type 5  passwords in a  file that  can then  be run\n");
		printf("    through john-the-ripper.\n\n");
		printf("    %s--dictionary=<file>%s\n", COL_GREEN, RESET);
		printf("    Use a dictionary  file to  check the  passwords  and  keys  are  not\n");
		printf("    dictionary-based  words.  If this  option is  not specified  a small\n");
		printf("    number of  built-in  passwords  are  used  instead.  This option  is\n");
		printf("    not used  to break  MD5 passwords,  only compare  passwords  already\n");
		printf("    known.\n\n");
		printf("    %s--pass-length=<length>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords have\n");
		printf("    a minimum length as specified. The default is %d characters.\n\n", nipper.passwordLength);
		printf("    %s--pass-uppers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords must\n");
		printf("    include uppercase characters. The default is ");
		if (nipper.passwordUppers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-lowers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords must\n");
		printf("    include lowercase characters. The default is ");
		if (nipper.passwordLowers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-either=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords must\n");
		printf("    include upper or lowercase characters. The default is ");
		if (nipper.passwordEither == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-numbers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords must\n");
		printf("    include numbers. The default is ");
		if (nipper.passwordNumbers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-specials=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength  checks will check  that retrieved  passwords must\n");
		printf("    include special characters. The default is ");
		if (nipper.passwordSpecials == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");

		printf("%sSecurity Audit Misc Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--timeout=<seconds>%s\n", COL_GREEN, RESET);
		printf("    Specify the  number of  seconds that  connection timeouts  should be\n");
		printf("    less than or equal to.  If timeouts are  less than this,  a security\n");
		printf("    will be generated in the report. The default is %d minutes.\n\n", (nipper.connectionTimeout / 60));

#if !defined(__win32__)
		printf("%sRemote SNMP Config Retrival Options:%s\n\n", COL_BLUE, RESET);
		printf("    These options allow the retrieval of a Cisco configuration file from\n");
		printf("    the device itself.  There are two  methods of  remote retrival;  the\n");
		printf("    first uses the old  Cisco system  MIB which  works with devices upto\n");
		printf("    IOS 12.x. The second  method uses the new Cisco  config copy MIB and\n");
		printf("    works  with  devices  running  IOS 12.0  (some  with  11.2P). Nipper\n");
		printf("    defaults to using the new Cisco system MIB.\n");
		printf("    This  feature  requires  a  running  TFTP  server  and the  net-snmp\n");
		printf("    package. Additionally, the  IP address of the  Cisco device  and the\n");
		printf("    local address  need to  be specified.  You will need  to be the root\n");
	    printf("    user.\n\n");
		printf("    %s--cisco-ip=<ip address>%s\n", COL_GREEN, RESET);
		printf("    Specify the  IP address  of the Cisco Device  to retrieve the config\n");
		printf("    from.\n\n");
		printf("    %s--local-ip=<ip address>%s\n", COL_GREEN, RESET);
		printf("    The IP address of the host making the request. This is essential for\n");
		printf("    remote config retrival, else it will fail.\n\n");
		printf("    %s--cisco=<old | new>%s\n", COL_GREEN, RESET);
		printf("    There are two methods of retrieving  the config remotely using SNMP.\n");
		printf("    The old  method is  supported on  Cisco devices  upto  IOS 12.x, the\n");
		printf("    new method  (default) is  supported from  IOS 12.0  (11.2P  on  some\n");
		printf("    devices).\n\n");
		printf("    %s--snmp=<community>%s\n", COL_GREEN, RESET);
		printf("    The snmp community to be used with the above. A default of 'private'\n");
		printf("    will be used if not specified.\n\n");
		printf("    %s--tftproot=<directory>%s\n", COL_GREEN, RESET);
		printf("    The tftp root directory to be used for the Cisco config transfer. If\n");
		printf("    not specified a default of /tftproot/ will be used.\n\n");
		printf("    %s--cisco-file=<filename>%s\n", COL_GREEN, RESET);
		printf("    Specifies a  filename for the Cisco config file to  be saved  to. If\n");
		printf("    not specified,  then the  file will be  left in  the  tftproot  as a\n");
		printf("    random file name.\n\n");
#endif

		printf("%sDebug Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--debug%s\n", COL_GREEN, RESET);
		printf("    Outputs additional  information that most users  would not find very\n");
		printf("    helpful, but  is useful  in understanding  how the  configuration is\n");
		printf("    interpreted. Using the debug option overrides all other output.\n\n");
		printf("    %s--debug-abbreviations%s\n", COL_GREEN, RESET);
		printf("    Shows a list of all the  abbreviations that were  not processed. The\n");
		printf("    list is output  to stdout  and lists the  number of hits  each entry\n");
		printf("    has had. Not all of the items listed will be valid abbreviations.\n\n");
		printf("    %s--debug-lnp%s\n", COL_GREEN, RESET);
		printf("    Output the configuration lines not processed (to screen only).\n\n");

		printf("%sExamples:%s\n\n", COL_BLUE, RESET);
		printf("    %s%s --pix --input=configIOS.txt --output=report.html%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses the PIX configuration file configPIX.txt and outputs the HTML\n");
		printf("    report to the report.html file.\n\n");
		printf("    %scat ios.config | %s --company-name=Acme >report.html%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses  the Cisco  router  configuration  stored in  ios.config  and\n");
		printf("    redirects the  HTML report output  to report.html.  Nipper will also\n");
		printf("    use the company name \"Acme\" in the report output.\n\n");
		printf("    %s%s --pass-length=12 --input=cisco.txt --output=report.htm%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses the configuration  file stored  in cisco.txt and  outputs the\n");
		printf("    report to report.htm.  It will override the  default password length\n");
		printf("    check of %d and checks for a miniumum password length of 12 chars.\n\n", nipper.passwordLength);
	}

	// Show version?
	else if (showVersion == true)
	{
		printf("%s\n\n", program_version_banner);
	}


// ----------------------------------------------------
// PART 2: Process Configuration File

	else
	{
		// Debug Output?
		if (nipper.debugMode == true)
		{
			printf("%s\n", program_banner);

			printf("\n%sCommand Line Options\n====================%s\nCommand:", COL_BLUE, RESET);
			for (argLoop = 0; argLoop < argc; argLoop++)
			{
				printf(" %s", argv[argLoop]);
			}
			printf("\n\n");
		}

		// Create relevent device structure
		switch (nipper.deviceType)
		{
			// IOS-based...
			case type_ios_router:
			case type_ios_switch:
			case type_ios_catalyst:
				nipper.ios = malloc(sizeof(struct ciscoIOSConfig));
				memset(nipper.ios, 0, sizeof(struct ciscoIOSConfig));
				nipper.ios->sourceRouting = true;
				nipper.ios->tcpSmallServers = unconfigured;
				nipper.ios->udpSmallServers = unconfigured;
				nipper.ios->finger = unconfigured;
				nipper.ios->cdp = true;
				nipper.ios->cef = false;
				nipper.ios->servicePasswordEncryption = false;
				nipper.ios->bootp = true;
				nipper.ios->bootNetwork = false;
				nipper.ios->serviceConfig = false;
				nipper.ios->pad = true;
				nipper.ios->domainLookup = true;
				nipper.ios->classlessRouting = true;
				nipper.ios->keepAlivesIn = false;
				nipper.ios->keepAlivesOut = false;
				nipper.ios->gratuitousArps = false;
				nipper.ios->calendarValid = false;
				nipper.ios->passwordMinLength = -1;
				nipper.ios->vtpMode = vtp_mode_server;
				strcpy(nipper.ios->timeZone, "UTC");
				break;

			// NMP-based...
			case type_nmp_catalyst:
			case type_cos_catalyst:
				nipper.nmp = malloc(sizeof(struct ciscoNMPConfig));
				memset(nipper.nmp, 0, sizeof(struct ciscoNMPConfig));
				nipper.nmp->cdp = true;
				nipper.nmp->cdpInterval = 60;
				nipper.nmp->cdpHoldtime = 180;
				nipper.nmp->cdpVersion = 2;
				nipper.nmp->timeout = 20;
				nipper.nmp->portSecurityAuto = false;
				break;

			// PIX-based...
			case type_pix_firewall:
			case type_asa_firewall:
			case type_fwsm_firewall:
				nipper.pix = malloc(sizeof(struct ciscoPIXConfig));
				memset(nipper.pix, 0, sizeof(struct ciscoPIXConfig));
				nipper.pix->transparent = false;
				nipper.pix->accessListAlertInterval = 300;
				nipper.pix->accessListDenyFlowMax = 4096;
				break;

			case type_css_filter:
				nipper.css = malloc(sizeof(struct ciscoCSSConfig));
				memset(nipper.css, 0, sizeof(struct ciscoCSSConfig));
				nipper.css->snmpAccess = true;
				nipper.css->ftpAccess = true;
				nipper.css->telnetAccess = true;
				nipper.css->sshAccess = true;
				nipper.css->consoleAccess = true;
				nipper.css->webManagementAccess = false;
				nipper.css->aclEnabled = true;
				break;

			// SOS-based...
			case type_sos_firewall:
				nipper.sos = malloc(sizeof(struct screenOSConfig));
				memset(nipper.sos, 0, sizeof(struct screenOSConfig));
				nipper.sos->sshEnabled = false;
				nipper.sos->sshKeyGenTime = 60;
				break;
		}

#if !defined(__win32__)
		// Get remote config?
		if ((nipper.remoteIP[0] != 0) && (nipper.localIP[0] != 0) && (getuid() == 0))
		{

			// Create tftp file
			tempInt = rand();
			if (nipper.remoteTftp[strlen(nipper.remoteTftp) -1] != '/')
				nipper.remoteTftp[strlen(nipper.remoteTftp)] = '/';
			sprintf(tempCommand, "touch %s%d.config", nipper.remoteTftp, tempInt);
			system(tempCommand);

			// Set tftp file writable
			sprintf(tempCommand, "chmod 666 %s%d.config", nipper.remoteTftp, tempInt);
			system(tempCommand);

			// Old method?
			if (nipper.oldCisco == true)
			{
				// http://www.cisco.com/en/US/tech/tk648/tk362/technologies_tech_note09186a008009463e.shtml
				// Write to SNMP server
				sprintf(tempCommand, "snmpset -c %s -v 1 %s 1.3.6.1.4.1.9.2.1.55.%s s \"%d.config\"", nipper.remoteSnmp, nipper.remoteIP, nipper.localIP, tempInt);
				if (nipper.debugMode == true)
					printf("SNMP Command: %s\n", tempCommand);
				system(tempCommand);
			}

			// new method
			else
			{
				// http://www.cisco.com/en/US/tech/tk648/tk362/technologies_tech_note09186a0080094aa6.shtml
				// Write to SNMP server
				tempInt2 = rand();
				sprintf(tempCommand, "snmpset -c %s -v 1 %s .1.3.6.1.4.1.9.9.96.1.1.1.1.2.%d i 1 .1.3.6.1.4.1.9.9.96.1.1.1.1.3.%d i 4 .1.3.6.1.4.1.9.9.96.1.1.1.1.4.%d i 1 .1.3.6.1.4.1.9.9.96.1.1.1.1.5.%d a \"%s\" .1.3.6.1.4.1.9.9.96.1.1.1.1.6.%d s \"%d.config\" .1.3.6.1.4.1.9.9.96.1.1.1.1.14.%d i 4", nipper.remoteSnmp, nipper.remoteIP, tempInt2, tempInt2, tempInt2, tempInt2, nipper.localIP, tempInt2, tempInt, tempInt2);
				if (nipper.debugMode == true)
					printf("SNMP Command: %s\n", tempCommand);
				system(tempCommand);
			}

			// Wait for a couple of secs (for file to arrive)
			system("sleep 3");

			// Check the file has content
			fileStats = malloc(sizeof(struct stat));
			memset(tempCommand, 0, sizeof(tempCommand));
			sprintf(tempCommand, "%s%d.config", nipper.remoteTftp, tempInt);
			stat(tempCommand, fileStats);

			// Cleanup SNMP if new...
			if (nipper.oldCisco == false)
			{
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "snmpset -c %s -v 1 %s .1.3.6.1.4.1.9.9.96.1.1.1.1.14.%d i 6", nipper.remoteSnmp, nipper.remoteIP, tempInt2);
				system(tempCommand);
			}

			// Exit if file has nothing in it...
			if (fileStats->st_size == 0)
			{
				free(fileStats);
				unlink(tempCommand);
				printf("\n%sERROR: Could not retrieve the config remotely.%s\n", COL_RED, RESET);
				return 1;
			}
			free(fileStats);

			// If file is to called something...
			if (nipper.localSave != 0)
			{
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "mv %s%d.config %s", nipper.remoteTftp, tempInt, argv[nipper.localSave]+13);
				system(tempCommand);
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "%s", argv[nipper.localSave]+13);
			}

			// Open file as input
			nipper.input = fopen(tempCommand, "r");
			if (nipper.input == NULL)
			{
				printf("\n%sERROR: Could not open the remote config file.%s\n", COL_RED, RESET);
				return 1;
			}
		}
		else if ((nipper.remoteIP[0] != 0) && (nipper.localIP[0] != 0) && (getuid() != 0))
		{
			printf("\n%sERROR: You must be the root user to get the Cisco config remotely.%s\n", COL_RED, RESET);
			return 1;
		}
		else if ((nipper.remoteIP[0] != 0) || (nipper.localIP[0] != 0))
		{
			printf("\n%sERROR: You must be the specify Cisco device and local IP.%s\n", COL_RED, RESET);
			return 1;
		}

		// From Stdin...
		else if (inputFile == 0)
#else
		if (inputFile == 0)
#endif
			nipper.input = stdin;

		// From file...
		else
		{
			// Check the file has some contents...
			fileStats = malloc(sizeof(struct stat));
			stat(argv[inputFile]+8, fileStats);
			if (fileStats->st_size == 0)
			{
				free(fileStats);
				printf("\n%sERROR: The configuration file is empty!.%s\n", COL_RED, RESET);
				return 1;
			}
			free(fileStats);

			nipper.input = fopen(argv[inputFile]+8, "r");
			if (nipper.input == NULL)
			{
				printf("\n%sERROR: Could not open input file '%s'.%s\n", COL_RED, argv[inputFile]+8, RESET);
				return 1;
			}
		}

		if (nipper.debugMode == true)
		{
			printf("\n%sConfig Lines Examined\n=====================%s\n", COL_BLUE, RESET);
		}

		// Read in input and process...
		while (feof(nipper.input) == 0)
		{
			// Read line
			readLine(nipper.input, line, sizeof(line));

			// Split the command line up
			command = splitLine(line);

			switch (nipper.deviceType)
			{
				// IOS devices
				case type_ios_router:
				case type_ios_switch:
				case type_ios_catalyst:

					// Init
					if (strcmp(command.part[0], "no") == 0)
					{
						tempInt = 1;
						setting = false;
					}
					else
					{
						tempInt = 0;
						setting = true;
					}

					if (strncmp("!", line, 1) == 0)
					{ }

					// Hostname
					else if (strcmp(command.part[tempInt], "hostname") == 0)
					{
						if (nipper.debugMode == true)
							printf("Hostname Line: %s\n", line);
						strncpy(nipper.hostname, command.part[tempInt+1], sizeof(nipper.hostname) - 1);
					}

					// Version
					else if (strcmp(command.part[tempInt], "version") == 0)
					{
						if (nipper.debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(nipper.version, command.part[tempInt+1], sizeof(nipper.version) - 1);
						nipper.versionMajor = atoi(nipper.version);
						nipper.versionMinor = atoi(strchr(nipper.version, '.') + 1);
					}

					// Service lines...
					else if (strcmp(command.part[tempInt], "service") == 0)
					{
						tempInt++;

						// Service Password Encryption...
						if (strcmp(command.part[tempInt], "password-encryption") == 0)
						{
							if (nipper.debugMode == true)
								printf("Service Password-Encryption Line: %s\n", line);
							nipper.ios->servicePasswordEncryption = setting;
						}

						// TCP keep alives in
						else if (strcmp(command.part[tempInt], "tcp-keepalives-in") == 0)
						{
							if (nipper.debugMode == true)
								printf("TCP Keep Alives In Line: %s\n", line);
							nipper.ios->keepAlivesIn = setting;
						}

						// TCP keep alives out
						else if (strcmp(command.part[tempInt], "tcp-keepalives-out") == 0)
						{
							if (nipper.debugMode == true)
								printf("TCP Keep Alives Out Line: %s\n", line);
							nipper.ios->keepAlivesOut = setting;
						}

						// Service Config
						else if (strcmp(command.part[tempInt], "config") == 0)
							processServiceConfig(line, &nipper);

						// PAD Service...
						else if (strcmp(command.part[tempInt], "pad") == 0)
						{
							if (nipper.debugMode == true)
								printf("PAD Line: %s\n", line);
							nipper.ios->pad = setting;
						}

						// TCP Small Servers...
						else if (strcmp(command.part[tempInt], "tcp-small-servers") == 0)
							processSmallServers(line, &nipper);

						// UDP Small Servers...
						else if (strcmp(command.part[tempInt], "udp-small-servers") == 0)
							processSmallServers(line, &nipper);

						// Finger...
						else if (strcmp(command.part[tempInt], "finger") == 0)
						{
							if (nipper.debugMode == true)
								printf("Finger Line: %s\n", line);
							nipper.ios->finger = setting;
						}

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}

					// IP lines...
					else if (strcmp(command.part[tempInt], "ip") == 0)
					{
						tempInt++;

						// Cisco Express Forwarding
						if (strcmp(command.part[tempInt], "cef") == 0)
							processCEF(line, &nipper);

						// IP Classless Routing
						else if (strcmp(command.part[tempInt], "classless") == 0)
						{
							if (nipper.debugMode == true)
								printf("Classless Routing Line: %s\n", line);
							nipper.ios->classlessRouting = setting;
						}

						// Access Lists...
						else if (strcmp(command.part[tempInt], "access-list") == 0)
							processIPAccessList(line, &nipper);

						// Gratuitous ARPs
						else if (strcmp(command.part[tempInt], "gratuitous-arps") == 0)
						{
							if (nipper.debugMode == true)
								printf("Gratuitous ARPs Line: %s\n", line);
							nipper.ios->gratuitousArps = setting;
						}

						// SSH
						else if (strcmp(command.part[tempInt], "ssh") == 0)
							processSsh(line, &nipper);

						// Name Servers
						else if (strcmp(command.part[tempInt], "name-server") == 0)
							processNameServer(line, &nipper);

						// Domain Name
						else if (strcmp(command.part[tempInt], "domain-name") == 0)
						{
							if (nipper.debugMode == true)
								printf("Domain Name Line: %s\n", line);
							strncpy(nipper.domainName, command.part[tempInt+1], sizeof(nipper.domainName) - 1);
						}

						// Domain Lookup
						else if (strcmp(command.part[tempInt], "domain-lookup") == 0)
							processDomainLookup(line, &nipper);

						// Source routing...
						else if (strcmp(command.part[tempInt], "source-route") == 0)
						{
							if (nipper.debugMode == true)
								printf("IP Source Route Line: %s\n", line);
							nipper.ios->sourceRouting = setting;
						}

						// Bootp...
						else if ((strcmp(command.part[tempInt], "bootp") == 0) && (strcmp(command.part[tempInt+1], "server") == 0))
							processBootp(line, &nipper);

						// HTTP...
						else if (strcmp(command.part[tempInt], "http") == 0)
							processHttp(line, &nipper);

						// FTP username / password...
						else if (strcmp(command.part[tempInt], "ftp") == 0)
							processFtp(line, &nipper);

						// IP Route...
						else if ((strcmp(command.part[tempInt], "route") == 0) && (strcmp(command.part[tempInt+1], "vrf") != 0))
							processRoute(line, &nipper);

						// IP NAT...
						else if (strcmp(command.part[tempInt], "nat") == 0)
							processNAT(line, &nipper);

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}

					// Security
					else if (strcmp(command.part[tempInt], "security") == 0)
					{
						tempInt++;

						// Passwords Minimum Length
						if ((strcmp(command.part[tempInt], "passwords") == 0) && (strcmp(command.part[tempInt+1], "min-length") == 0))
						{
							if (nipper.debugMode == true)
								printf("Password Length Line: %s\n", line);
							if (setting == false)
								nipper.ios->passwordMinLength = 0;
							else
								nipper.ios->passwordMinLength = atoi(command.part[tempInt+2]);
						}

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}

					// Enable Passwords
					else if (strcmp(command.part[tempInt], "enable") == 0)
						processEnable(line, &nipper);

					// Usernames...
					else if (strcmp(command.part[tempInt], "username") == 0)
						processUsername(line, &nipper);

					// Interfaces
					else if (strcmp(command.part[tempInt], "interface") == 0)
						processInterface(line, &nipper);

					// Banner...
					else if (strcmp(command.part[tempInt], "banner") == 0)
						processBanner(line, &nipper);

					// CDP...
					else if ((strcmp(command.part[tempInt], "cdp") == 0) && (strcmp(command.part[tempInt+1], "run") == 0))
					{
						if (nipper.debugMode == true)
							printf("CDP Line: %s\n", line);
						nipper.ios->cdp = setting;
					}

					// Access Lists...
					else if (strcmp(command.part[tempInt], "access-list") == 0)
						processAccessList(line, &nipper);

					// Line...
					else if (strcmp(command.part[tempInt], "line") == 0)
						processLine(line, &nipper);

					// VTP...
					else if (strcmp(command.part[tempInt], "vtp") == 0)
						processVTP(line, &nipper);

					// Key chain...
					else if ((strcmp(command.part[tempInt], "key") == 0) && (strcmp(command.part[tempInt+1], "chain") == 0))
						processKeyChain(line, &nipper);

					// Routing...
					else if (strcmp(command.part[tempInt], "router") == 0)
						processRouter(line, &nipper);

					// Route Map...
					else if (strcmp(command.part[tempInt], "route-map") == 0)
						processRouteMap(line, &nipper);

					// TACACS...
					else if (strcmp(command.part[tempInt], "tacacs-server") == 0)
						processTacacs(line, &nipper);

					// AAA...
					else if (strcmp(command.part[tempInt], "aaa") == 0)
						processAaa(line, &nipper);

					// NTP...
					else if (strcmp(command.part[tempInt], "ntp") == 0)
						processNtp(line, &nipper);

					// Privilage Levels...
					else if ((strcmp(command.part[tempInt], "privilage") == 0) && (strcmp(command.part[tempInt+1], "exec") == 0) && (strcmp(command.part[tempInt+2], "level") == 0))
						processPrivilage(line, &nipper);

					// SNMP...
					else if (strcmp(command.part[tempInt], "snmp-server") == 0)
						processSnmp(line, &nipper);

					// Logging...
					else if (strcmp(command.part[tempInt], "logging") == 0)
						processLogging(line, &nipper);

					// Clock...
					else if (strcmp(command.part[tempInt], "clock") == 0)
						processClock(line, &nipper);

					// Boot Network
					else if ((strcmp(command.part[tempInt], "boot") == 0) && (strcmp(command.part[tempInt+1], "network") == 0))
						processBootNetwork(line, &nipper);

					// Debug (lines not processed)
					else if (nipper.linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// NMP devices
				case type_nmp_catalyst:
				case type_cos_catalyst:

					// Init
					if (strcmp(command.part[0], "clear") == 0)
					{
						tempInt = 1;
						setting = false;
					}
					else if (strcmp(command.part[0], "set") == 0)
					{
						tempInt = 1;
						setting = true;
					}
					else
						tempInt = 0;

					// System lines...
					if (strcmp(command.part[tempInt], "system") == 0)
					{
						tempInt++;

						// Hostname
						if (strcmp(command.part[tempInt], "name") == 0)
						{
							if (nipper.debugMode == true)
								printf("Hostname Line: %s\n", line);
							strncpy(nipper.hostname, command.part[tempInt+1], sizeof(nipper.hostname) - 1);
						}

						// Location...
						else if (strcmp(command.part[tempInt], "location") == 0)
						{
							if (nipper.debugMode == true)
								printf("Location Line: %s\n", line);
							tempInt = 20;
							while (line[tempInt] == ' ')
								tempInt++;
							strncpy(nipper.nmp->location, line + tempInt, sizeof(nipper.nmp->location) - 1);
						}
	
						// Contact...
						else if (strcmp(command.part[tempInt], "contact") == 0)
						{
							if (nipper.debugMode == true)
								printf("Contact Line: %s\n", line);
							tempInt = 19;
							while (line[tempInt] == ' ')
								tempInt++;
							strncpy(nipper.nmp->contact, line + tempInt, sizeof(nipper.nmp->contact) - 1);
						}

						// Corefile...
						else if (strcmp(command.part[tempInt], "core-file") == 0)
						{
							if (nipper.debugMode == true)
								printf("Core-file Line: %s\n", line);
							strncpy(nipper.nmp->coreFile, command.part[tempInt+1], sizeof(nipper.nmp->coreFile) - 1);
						}

						// Syslog file...
						else if (strcmp(command.part[tempInt], "syslog-file") == 0)
						{
							if (nipper.debugMode == true)
								printf("Syslog-file Line: %s\n", line);
							strncpy(nipper.nmp->syslogFile, command.part[tempInt+1], sizeof(nipper.nmp->syslogFile) - 1);
						}

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}

					// IP lines...
					else if (strcmp(command.part[tempInt], "ip") == 0)
					{
						tempInt++;

						// Permit List
						if (strcmp(command.part[tempInt], "permit") == 0)
							processNMPPermit(line, &nipper);

						// ICMP IP Unreachables
						else if (strcmp(command.part[tempInt], "unreachable") == 0)
							processNMPUnreachable(line, &nipper);

						// IP Redirects
						else if (strcmp(command.part[tempInt], "redirect") == 0)
							processNMPRedirect(line, &nipper);

						// IP Fragmentation
						else if (strcmp(command.part[tempInt], "fragmentation") == 0)
							processNMPFragmentation(line, &nipper);

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}

					// Version
					else if (strcmp(command.part[tempInt], "#version") == 0)
					{
						if (nipper.debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(nipper.version, command.part[tempInt+1], sizeof(nipper.version) - 1);
						nipper.versionMajor = atoi(nipper.version);
						nipper.versionMinor = atoi(strchr(nipper.version, '.') + 1);
						nipper.versionRevision = atoi(strchr(nipper.version, '(') + 1);
					}

					// Logout Timeout...
					else if (strcmp(command.part[tempInt], "logout") == 0)
					{
						if (nipper.debugMode == true)
							printf("Logout Line: %s\n", line);
						nipper.nmp->timeout = atoi(command.part[tempInt+1]);
					}

					// Enable Passwords
					else if ((strcmp(command.part[tempInt], "password") == 0) || (strcmp(command.part[tempInt], "enablepass") == 0))
						processEnablePass(line, &nipper);

					// CDP...
					else if (strcmp(command.part[tempInt], "cdp") == 0)
						processNMPCDP(line, &nipper);

					// SNMP...
					else if (strcmp(command.part[tempInt], "snmp") == 0)
						processNMPSnmp(line, &nipper);

					// Ports
					else if (strcmp(command.part[tempInt], "port") == 0)
						processNMPPort(line, &nipper);

					// Trunk
					else if (strcmp(command.part[tempInt], "trunk") == 0)
						processNMPTrunk(line, &nipper);

					// Spanning Tree
					else if (strcmp(command.part[tempInt], "spantree") == 0)
						processNMPSpantree(line, &nipper);

					// VLAN
					else if (strcmp(command.part[tempInt], "vlan") == 0)
						processNMPVlan(line, &nipper);

					// Debug (lines not processed)
					else if (nipper.linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// PIX devices
				case type_pix_firewall:
				case type_asa_firewall:
				case type_fwsm_firewall:

					// Hostname
					if (strcmp(command.part[0], "hostname") == 0)
					{
						if (nipper.debugMode == true)
							printf("Hostname Line: %s\n", line);
						strncpy(nipper.hostname, command.part[1], sizeof(nipper.hostname) - 1);
					}

					// Version
					else if (((strcmp(command.part[0], "PIX") == 0) || (strcmp(command.part[0], "ASA") == 0) || (strcmp(command.part[0], "FWSM") == 0)) && (strcmp(command.part[1], "Version") == 0))
					{
						if (nipper.debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(nipper.version, command.part[2], sizeof(nipper.version) - 1);
						nipper.versionMajor = atoi(nipper.version);
						nipper.versionMinor = atoi(strchr(nipper.version, '.') + 1);
						nipper.versionRevision = atoi(strchr(nipper.version, '(') + 1);
						if (strcmp(command.part[0], "PIX") == 0)
							nipper.deviceType = type_pix_firewall;
						else if (strcmp(command.part[0], "ASA") == 0)
							nipper.deviceType = type_asa_firewall;
						else if (strcmp(command.part[0], "FWSM") == 0)
							nipper.deviceType = type_fwsm_firewall;
					}

					// Transparent Firewall?
					else if ((strcmp(command.part[0], "firewall") == 0) && (strcmp(command.part[1], "transparent") == 0))
					{
						if (nipper.debugMode == true)
							printf("Transparent Line: %s\n", line);
						nipper.pix->transparent = true;
					}

					// ICMP Access...
					else if (strcmp(command.part[0], "icmp") == 0)
						processICMPPIX(line, &nipper);

					// Enable Password
					else if (((strcmp(command.part[0], "enable") == 0) && (strcmp(command.part[1], "password") == 0)) || (strcmp(command.part[0], "passwd") == 0))
						processEnablePIX(line, &nipper);

					// Usernames...
					else if (strcmp(command.part[0], "username") == 0)
						processPIXUsername(line, &nipper);

					// Route...
					else if (strcmp(command.part[0], "route") == 0)
						processRoutePIX(line, &nipper);

					// Interfaces...
					else if ((strcmp(command.part[0], "interface") == 0) || (strcmp(command.part[0], "nameif") == 0) || (strcmp(command.part[0], "mtu") == 0) || ((strcmp(command.part[0], "ip") == 0) && (strcmp(command.part[1], "address") == 0)) || (strcmp(command.part[0], "access-group") == 0))
						processPIXInterface(line, &nipper);

					// Object Group...
					else if (strcmp(command.part[0], "object-group") == 0)
						processObjectGroup(line, &nipper);

					// Protocol Inspection...
					else if (((strcmp(command.part[0], "fixup") == 0) && (strcmp(command.part[1], "protocol") == 0)) || ((strcmp(command.part[0], "no") == 0) && (strcmp(command.part[1], "fixup") == 0) && (strcmp(command.part[2], "protocol") == 0)))
						processFixupPIX(line, &nipper);

					// Name...
					else if (strcmp(command.part[0], "name") == 0)
						processPIXName(line, &nipper);

					// Domain Name
					else if (strcmp(command.part[0], "domain-name") == 0)
						strncpy(nipper.domainName, line + 12, sizeof(nipper.domainName) - 1);

					// Access Lists...
					else if (strcmp(command.part[0], "access-list") == 0)
						processPIXAccessList(line, &nipper);

					// Static NAT / PAT
					else if (strcmp(command.part[0], "static") == 0)
						processStaticPIX(line, &nipper);

					// SSH
					else if (strcmp(command.part[0], "ssh") == 0)
						processPIXSSH(line, &nipper);

					// Debug (lines not processed)
					else if (nipper.linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// Cisco CSS devices
				case type_css_filter:
					// No line?
					if (strcmp(command.part[0], "no") == 0)
						tempInt = 1;
					else
						tempInt = 0;

					// Version
					if ((strcmp(command.part[0], "!Active") == 0) && (strcmp(command.part[1], "version:") == 0))
					{
						if (nipper.debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(nipper.version, command.part[2] + 2, 2);
						nipper.versionMajor = atoi(nipper.version);
						strncpy(nipper.version, command.part[2] + 4, 2);
						nipper.versionMinor = atoi(nipper.version);
						strncpy(nipper.version, command.part[2] + 6, 3);
						nipper.versionRevision = atoi(nipper.version);
						sprintf(nipper.version, "%d.%d(%d)", nipper.versionMajor, nipper.versionMinor, nipper.versionRevision);
					}

					// Name (Config doesn't include a name, this is a little bit of a hack.)
					else if ((nipper.hostname[0] == 0) && (strncmp(command.part[0] + strlen(command.part[0]) - 1, "#", 1) == 0) && (strncmp(command.part[1], "sh", 2) == 0) && (strncmp(command.part[2], "run", 3) == 0))
					{
						if (nipper.debugMode == true)
							printf("Hostname Line: %s\n", line);
						strncpy(nipper.hostname, command.part[0], sizeof(nipper.hostname) - 1);
						nipper.hostname[strlen(nipper.hostname) - 1] = 0;
					}

					// Restrict
					else if (strcmp(command.part[tempInt], "restrict") == 0)
						processCSSRestrict(line, &nipper);

					// SNMP
					else if (strcmp(command.part[tempInt], "snmp") == 0)
						processCSSSNMP(line, &nipper);

					// ACL
					else if (strcmp(command.part[tempInt], "acl") == 0)
						processCSSACL(line, &nipper);

					// Debug (lines not processed)
					else if (nipper.linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// ScreenOS devices
				case type_sos_firewall:

					if ((strcmp(command.part[0], "set") == 0) || (strcmp(command.part[0], "unset") == 0))
					{
						// Hostname
						if (strcmp(command.part[1], "hostname") == 0)
						{
							if (nipper.debugMode == true)
								printf("Hostname Line: %s\n", line);
							stripQuotes(command.part[2], line, sizeof(line));
							strncpy(nipper.hostname, line, sizeof(nipper.hostname) - 1);
						}

						// Admin
						else if (strcmp(command.part[1], "admin") == 0)
							processSOSAdmin(line, &nipper);

						// Auth Server
						else if (strcmp(command.part[1], "auth-server") == 0)
							processSOSAuthServer(line, &nipper);

						// Addresses
						else if (strcmp(command.part[1], "address") == 0)
							processSOSAddress(line, &nipper);

						// Interface
						else if (strcmp(command.part[1], "interface") == 0)
							processSOSInterface(line, &nipper);

						// Policy
						else if (strcmp(command.part[1], "policy") == 0)
							processSOSPolicy(line, &nipper);

						// SSH
						else if (strcmp(command.part[1], "ssh") == 0)
							processSOSSSH(line, &nipper);

						// Debug (lines not processed)
						else if (nipper.linesnotprocessed == true)
							printf("%s\n", line);
					}
					break;
			}
		}

		// Close input file...
		fclose(nipper.input);

		// Check that version has been identified for relevent configs...
		showUsage = false;
		switch (nipper.deviceType)
		{
			case type_ios_switch:
			case type_ios_router:
			case type_ios_catalyst:
				if (nipper.versionMajor == 0)
					showUsage = true;
				break;

			case type_nmp_catalyst:
			case type_cos_catalyst:
				if (nipper.versionMajor == 0)
					showUsage = true;
				break;

			case type_pix_firewall:
			case type_asa_firewall:
			case type_fwsm_firewall:
				if (nipper.versionMajor == 0)
					showUsage = true;
				break;

			case type_css_filter:
				if (nipper.versionMajor == 0)
					showUsage = true;
				break;

			case type_sos_firewall:
				if (nipper.hostname[0] == 0)
					showUsage = true;
				break;
		}
		if (showUsage == true)
		{
			printf("\n%sERROR: You are parsing the wrong type of configuration.\n       Specify the configuration type.%s\n", COL_RED, RESET);
			goto cleanup;
		}


// ----------------------------------------------------
// PART 3: Report Pre-parsing

		// Default Settings...
		processDefaults(&nipper);

		// Convert Names to IP Addresses...
		if (nipper.names == false)
			processNameMappings(&nipper);


// ----------------------------------------------------
// PART 4: Reporting

		// Output debug information?
		if (nipper.debugMode == true)
		{
			switch (nipper.deviceType)
			{
				case type_ios_router:
				case type_ios_switch:
				case type_ios_catalyst:
					reportIOSDebug(&nipper);
					break;

				case type_pix_firewall:
				case type_asa_firewall:
				case type_fwsm_firewall:
					reportPIXDebug(&nipper);
					break;

				case type_nmp_catalyst:
				case type_cos_catalyst:
					reportNMPDebug(&nipper);
					break;

				case type_css_filter:
					reportCSSDebug(&nipper);
					break;

				case type_sos_firewall:
					reportSOSDebug(&nipper);
					break;
			}
		}

		// Output report & john (if specified)
		else
		{
			// Output john?
			if ((nipper.johnFile[0] != 0) && (nipper.john != 0))
			{
				// Variables
				FILE *johnFile;

				// Create file
				johnFile = fopen(nipper.johnFile, "w");
				if (johnFile == NULL)
				{
					printf("\n%sERROR: Could not create john-the-ripper file '%s'.%s\n", COL_RED, nipper.johnFile, RESET);
					return 1;
				}

				// Write file contents
				johnPointer = nipper.john;
				while (johnPointer != 0)
				{
					fprintf(johnFile, "%s:%s\n", johnPointer->username, johnPointer->password);
					johnPointer = johnPointer->next;
				}

				// Close file
				fclose(johnFile);
			}

			// Generate the report
			switch (nipper.deviceType)
			{
				case type_ios_router:
				case type_ios_switch:
				case type_ios_catalyst:
					generateIOSReport(&nipper);
					break;

				case type_pix_firewall:
				case type_asa_firewall:
				case type_fwsm_firewall:
					generatePIXReport(&nipper);
					break;

				case type_cos_catalyst:
				case type_nmp_catalyst:
					generateNMPReport(&nipper);
					break;

				case type_css_filter:
					generateCSSReport(&nipper);
					break;

				case type_sos_firewall:
					generateSOSReport(&nipper);
					break;
			}

			// Output Debug Abbreviations
			if (nipper.debugAbbrev == true)
			{
				printf("\n%sUnknown Abbreviations\n=====================%s\n", COL_BLUE, RESET);
				debugAbPointer = debugAb;
				while (debugAbPointer != 0)
				{
					printf("  %s (%d)\n", debugAbPointer->ab, debugAbPointer->count);
					debugAbPointer = debugAbPointer->next;
				}
			}
		}


// ----------------------------------------------------
// PART 5: Cleanup

cleanup:

		// Debug Abbreviations
		while (debugAb != 0)
		{
			debugAbPointer = debugAb->next;
			free(debugAb);
			debugAb = debugAbPointer;
		}

		// Cleanup device type structures
		switch (nipper.deviceType)
		{
			case type_ios_router:
			case type_ios_switch:
			case type_ios_catalyst:
				cleanupIOS(&nipper);
				break;

			case type_pix_firewall:
			case type_asa_firewall:
			case type_fwsm_firewall:
				cleanupPIX(&nipper);
				break;

			case type_cos_catalyst:
			case type_nmp_catalyst:
				cleanupNMP(&nipper);
				break;

			case type_css_filter:
				cleanupCSS(&nipper);
				break;

			case type_sos_firewall:
				cleanupSOS(&nipper);
				break;
		}

		// Cleanup John-the-ripper
		while (nipper.john != 0)
		{
			johnPointer = nipper.john->next;
			free(nipper.john);
			nipper.john = johnPointer;
		}

		// close dict file
		if (nipper.dictionaryFile != 0)
		{
			fclose(nipper.dictionaryFile);
		}

		// close html css file
		if (nipper.cssFile != 0)
		{
			fclose(nipper.cssFile);
		}
	}

	return 0;
}
